import esphome.codegen as cg

resistance_sampler_ns = cg.esphome_ns.namespace("resistance_sampler")
ResistanceSampler = resistance_sampler_ns.class_("ResistanceSampler")

CODEOWNERS = ["@jesserockz"]
